@@ -13,6 +13,7 @@
 # limitations under the License.
 
 
+import pathlib
 from ._dcap_attestation import validate_attestation
 from .utils import *
 
@@ -24,6 +25,7 @@ import os
 import contextlib
 import socket
 
+import numpy as np
 import cbor2 as cbor
 
 from hashlib import sha256
@@ -55,6 +57,29 @@ class ModelDatumType(IntEnum):
     Bool = 10
 
 
+format_per_item = {
+    ModelDatumType.F32: "<f4",
+    ModelDatumType.F64: "<f8",
+    ModelDatumType.I32: "<i4",
+    ModelDatumType.I64: "<i8",
+    ModelDatumType.U32: "<u4",
+    ModelDatumType.U64: "<u8",
+    ModelDatumType.U8: "<u1",
+    ModelDatumType.U16: "<u2",
+    ModelDatumType.I8: "<i1",
+    ModelDatumType.I16: "<i2",
+    ModelDatumType.Bool: "?",
+}
+
+
+def serialize_tensor(tensor: np.ndarray, type: ModelDatumType) -> bytes:
+    return np.array(tensor).astype(format_per_item[type], casting="equiv").tobytes()
+
+
+def deserialize_tensor(data: bytes, type: ModelDatumType) -> np.ndarray:
+    return np.frombuffer(data, dtype=format_per_item[type])
+
+
 class TensorInfo:
     fact: List[int]
     datum_type: ModelDatumType
@@ -76,21 +101,20 @@ class Tensor:
     """
 
     info: Union[TensorInfo, dict]
-    bytes_data: List[int]
+    bytes_data: bytes
 
     def __init__(self, info: Union[TensorInfo, dict], bytes_data: bytes):
         self.info = info
-        self.bytes_data = list(bytes_data)
+        self.bytes_data = bytes_data
 
     def as_flat(self) -> list:
         """Convert the prediction calculated by the server to a flat python list."""
-        return cbor.loads(bytes(self.bytes_data))
+        return self.as_numpy().tolist()
 
     def as_numpy(self):
         """Convert the prediction calculated by the server to a numpy array."""
-        import numpy
 
-        arr = numpy.array([*self.as_flat()], dtype=dtype_to_numpy(self.info.datum_type))
+        arr = deserialize_tensor(self.bytes_data, self.info.datum_type)
         arr.shape = self.shape
         return arr
 
@@ -107,11 +131,7 @@ class Tensor:
             raise ImportError(
                 "torch not installed, please install with pip install blindai[torch]"
             ) from e
-
-        arr = torch.asarray(
-            [*self.as_flat()],
-            dtype=getattr(torch, dtype_to_torch(self.info.datum_type)),
-        )
+        arr = torch.tensor(self.as_numpy())
         return arr.view(self.shape)
 
     @property
@@ -365,11 +385,11 @@ def translate_tensor(
     """
     if _is_torch_tensor(tensor):
         info = TensorInfo(tensor.shape, translate_dtype(tensor.dtype), name)
-        iterable = tensor.flatten().tolist()
+        iterable = tensor.flatten().numpy()
 
     elif _is_numpy_array(tensor):
         info = TensorInfo(tensor.shape, translate_dtype(tensor.dtype), name)
-        iterable = tensor.flatten().tolist()
+        iterable = tensor.flatten()
 
     else:
         # Input is flatten tensor.
@@ -379,7 +399,7 @@ def translate_tensor(
             )
 
         info = TensorInfo(or_shape, translate_dtype(or_dtype), name)
-        iterable = tensor
+        iterable = np.array(tensor, dtype=dtype_to_numpy(or_dtype))
 
     if or_dtype is not None and or_dtype != info.datum_type:
         raise ValueError(
@@ -387,7 +407,7 @@ def translate_tensor(
         )
 
     # todo validate tensor content, dtype and shape
-    return Tensor(info.__dict__, cbor.dumps(iterable))
+    return Tensor(info.__dict__, serialize_tensor(iterable, info.datum_type))
 
 
 def translate_tensors(tensors, dtypes, shapes) -> List[dict]:
@@ -461,35 +481,24 @@ class BlindAiConnection(contextlib.AbstractContextManager):
     def __init__(
         self,
         addr: str,
-        server_name: str = "blindai-srv",
-        # manifest: Optional[str] = None,
-        certificate: Optional[str] = None,
-        simulation: bool = False,
         untrusted_port: int = 9923,
         attested_port: int = 9924,
+        hazmat_manifest_path: Optional[pathlib.Path] = None,
+        hazmat_http_on_untrusted_port=False,
     ):
         """
-        Connect to the server with the specified parameters.
-        You will have to specify here the expected manifest (server identity, configuration...)
-        and the server TLS certificate, if you are using the hardware mode.
-        If you're using the simulation mode, you don't need to provide a manifest and certificate,
-        but please keep in mind that this mode should NEVER be used in production as it doesn't
-        have most of the security provided by the hardware mode.
-        ***Security & confidentiality warnings:***
-           *manifest: Defines the rules upon which enclaves are accepted (after quote data verification). Contains the hash of MRENCLAVE which helps identify code and data of an enclave. In the case of leakeage of this file, data & model confidentiality would not be affected as the information just serves as a verification check.
-           For more details, the attestation info is verified against the manifest for the quote. In case of a leakage of the information of this file, code and data inside the secure enclave will remain inaccessible.
-           certificate:  The certificate file, which is also generated server side, is used to assigned the claims the manifest is checked against. It serves to identify the server for creating a secure channel and begin the attestation process.*
+        Connect to the BlindAi server
+
         Args:
-            addr (str): The address of BlindAI server you want to reach.
-            server_name (str, optional): Contains the CN expected by the server TLS certificate. Defaults to "blindai-srv".
-            manifest (Optional[str], optional): Path to the toml file describing the manifest of the server.
-                Generated in the server side. Defaults to None.
-                If left to none and if you are in hardware mode, the built-in manifest will be used.
-            certificate (Optional[str], optional): Path to the public key of the untrusted inference server.
-                Generated in the server side. Defaults to None.
-                If left to none and if you are in hardware mode, the certificate verification will be disabled
-            simulation (bool, optional): Connect to the server in simulation mode.
-                If set to True, the args manifest and certificate will be ignored. Defaults to False.
+            addr (str): The address of BlindAI server you want to connect to. It can be a domain (such as `example.com` or `localhost`) or an IP
+            hazmat_manifest_path: By default the built-in Manifest.toml provided by Mithril Security will be use.
+                You can override the default Manifest.toml by providing a path to your custom Manifest.toml
+                Caution: The manifest describes which enclave are trustworthy, changing the Manifest.toml can impact the security of the solution.
+            hazmat_http_on_untrusted_port: By default, the client fetch the attestation elements from the untrusted port of the server
+                using an HTTPS connection. The certificate should be validated according to your OS defaults.
+                You can opt out of a HTTPS connection and instead ask the client to connect via HTTP by setting this param to True
+                Caution: This parameter should never be set to True in production. Using a HTTPS connection is critical to
+                get a graceful degradation in case of a failure of Intel SGX attestation.
             untrusted_port (int, optional): Untrusted connection server port. Defaults to 9923.
             attested_port (int, optional): Attested connection server port. Defaults to 9924.
         Raises:
@@ -517,43 +526,33 @@ class BlindAiConnection(contextlib.AbstractContextManager):
 
         self._connect_server(
             addr,
-            server_name,
-            # manifest,
-            certificate,
-            simulation,
             untrusted_port,
             attested_port,
+            hazmat_manifest_path,
+            hazmat_http_on_untrusted_port,
         )
 
     def _connect_server(
         self,
         addr: str,
-        server_name,
-        # manifest,
-        certificate,
-        simulation,
         untrusted_port,
         attested_port,
+        manifest_path,
+        http_on_untrusted_port,
     ):
-        self.simulation_mode = simulation
-        self._disable_untrusted_server_cert_check = simulation
 
-        # addr = strip_https(addr)
+        if http_on_untrusted_port:
+            self._untrusted_url = "http://" + addr + ":" + str(untrusted_port)
+        else:
+            self._untrusted_url = "https://" + addr + ":" + str(untrusted_port)
 
-        self._untrusted_url = "http://" + addr + ":" + str(untrusted_port)
         self._attested_url = "https://" + addr + ":" + str(attested_port)
 
-        # if not self.simulation_mode:
-        #    self.manifest = manifest.from_file(manifest)
-
         # This adapter makes it possible to connect
         # to the server via a different hostname
         # that the one included in the certificate i.e. blindai-srv
         # For instance we can use it to connect to the server via the
         # domain / IP provided to connect(). See below
-
-        # TODO: Document that in production one should not use that approach
-        # but instead include the real domain name in the certificate
         class CustomHostNameCheckingAdapter(HTTPAdapter):
             def cert_verify(self, conn, url, verify, cert):
                 conn.assert_hostname = "blindai-srv"
@@ -565,29 +564,12 @@ class BlindAiConnection(contextlib.AbstractContextManager):
         # Always raise an exception when HTTP returns an error code for the untrusted connection
         # Note : we might want to do the same for the attested connection ?
         s.hooks = {"response": lambda r, *args, **kwargs: r.raise_for_status()}
-        # if self._disable_untrusted_server_cert_check:
-        #     logging.warning("Untrusted server certificate check bypassed")
-        #     s.verify = False
-        # else:
-        #     # verify is set to a path to the certificate CA
-        #     # It is used to pin the certificate
-        #     # Might not be needed in production
-        #     # Certificate pinning is a double edge sword
-        #     # It can prevent some MITM but it also make it more difficult
-        #     # to rotate the certificate...
-        #     # Anyway in our case as we'll do attestation verification
-        #     # it would make sense to use a simpler cert validation
-        #     # maybe simply the default CA with domain validation
-        #     # like the browser do.
-        #     # Can't be more boring (in a good way).
-        #     s.verify = certificate
-        # s.mount(self._untrusted_url, CustomHostNameCheckingAdapter())
 
         cert = cbor.loads(s.get(self._untrusted_url).content)
         quote = cbor.loads(s.get(f"{self._untrusted_url}/quote").content)
         collateral = cbor.loads(s.get(f"{self._untrusted_url}/collateral").content)
 
-        validate_attestation(quote, collateral, cert)
+        validate_attestation(quote, collateral, cert, manifest_path=manifest_path)
 
         # requests (http library) takes a path to a file containing the CA
         # there is no easy way to give the CA as a string/bytes directly
@@ -655,7 +637,7 @@ class BlindAiConnection(contextlib.AbstractContextManager):
     def run_model(
         self,
         model_id: str,
-        input_tensors: Optional[Union[List[List], Dict]] = None,
+        input_tensors: Optional[Union[List, Dict]] = None,
         dtypes: Optional[List[ModelDatumType]] = None,
         shapes: Optional[Union[List[List[int]], List[int]]] = None,
     ) -> RunModelResponse:
